import os
import yaml
import logging
from typing import Dict, List, Optional

from meltano.core.utils import nest
from .project import Project
from .plugin import Plugin, PluginType
from .plugin.factory import plugin_factory
from .plugin.error import PluginMissingError


class ConfigService:
    def __init__(self, project: Project):
        self.project = project

    def make_meltano_secret_dir(self):
        os.makedirs(self.project.meltano_dir(), exist_ok=True)

    def add_to_file(self, plugin: Plugin):
        with self.project.meltano_update() as meltano_yml:
            plugins = nest(meltano_yml, "plugins")
            plugins[plugin.type] = plugins.get(plugin.type, [])

        if plugin in self.plugins():
            logging.warning(
                f"{plugin.name} is already present, use `meltano install` to install it."
            )
            return

        with self.project.meltano_update() as meltano_yml:
            meltano_yml["plugins"][plugin.type].append(plugin.canonical())

    def has_plugin(self, plugin_name: str):
        try:
            self.get_plugin(plugin_name)
            return True
        except PluginMissingError:
            return False

    def get_plugin(self, plugin_name: str, plugin_type: Optional[PluginType] = None):
        try:
            return next(
                plugin
                for plugin in self.plugins()
                if (
                    plugin.name == plugin_name
                    and (plugin_type is None or plugin.type == plugin_type)
                )
            )
        except StopIteration:
            raise PluginMissingError(plugin_name)

    def get_extractors(self):
        return filter(lambda p: p.type == PluginType.EXTRACTORS, self.plugins())

    def get_loaders(self):
        return filter(lambda p: p.type == PluginType.LOADERS, self.plugins())

    def get_transformers(self):
        return filter(lambda p: p.type == PluginType.TRANSFORMERS, self.plugins())

    def get_transforms(self):
        return filter(lambda p: p.type == PluginType.TRANSFORMS, self.plugins())

    def get_models(self):
        return filter(lambda p: p.type == PluginType.MODELS, self.plugins())

    def get_database(self, database_name):
        return yaml.load(
            open(self.project.meltano_dir(f".database_{database_name}.yml"))
        )

    def plugins(self) -> List[Plugin]:
        """Parse the meltano.yml file and return it as `Plugin` instances."""
        # this will parse the meltano.yml file and create an instance of the
        # corresponding `plugin_class` for all the plugins.
        return (
            plugin_factory(plugin_type, plugin_def)
            for plugin_type, plugin_defs in self.project.meltano.get(
                "plugins", {}
            ).items()
            for plugin_def in plugin_defs
        )



import pytest
import hashlib
from flask import url_for


def test_get_default(api, app):
    with app.test_request_context():
        app_response = api.get(url_for("root.default"))
        res = api.get("/this/is__/a/__test")

        assert app_response.data == res.data



import json
import logging
from typing import Dict
from jsonschema import Draft4Validator

from meltano.core.utils import file_has_data
from meltano.core.behavior.hookable import hook
from meltano.core.plugin.error import PluginExecutionError

from . import SingerPlugin, PluginType
from .catalog import visit, SelectExecutor


class SingerTap(SingerPlugin):
    __plugin_type__ = PluginType.EXTRACTORS

    def exec_args(self, files: Dict):
        """
        Return the arguments list with the complete runtime paths.
        """
        args = ["--config", files["config"]]

        if file_has_data(files["catalog"]):
            args += ["--catalog", files["catalog"]]

        if file_has_data(files["state"]):
            args += ["--state", files["state"]]

        return args

    @property
    def config_files(self):
        return {
            "config": "tap.config.json",
            "catalog": "tap.properties.json",
            "state": "state.json",
        }

    @property
    def output_files(self):
        return {"output": "tap.out"}

    @hook("before_invoke", can_fail=True)
    def run_discovery(self, plugin_invoker, exec_args=[]):
        if not self._extras.get("autodiscover", True):
            return

        if "--discover" in exec_args:
            return

        properties_file = plugin_invoker.files["catalog"]

        with properties_file.open("w") as catalog:
            result = plugin_invoker.invoke("--discover", stdout=catalog)
            exit_code = result.wait()

        if exit_code != 0:
            properties_file.unlink()
            raise PluginExecutionError(
                f"Command {plugin_invoker.exec_path()} {plugin_invoker.exec_args()} returned {exit_code}"
            )

        # test for the schema to be a valid catalog
        try:
            with properties_file.open("r") as catalog:
                schema_valid = Draft4Validator.check_schema(json.load(catalog))
        except:
            logging.warn("Invalid catalog output by --discovery.")
            properties_file.unlink()

    @hook("before_invoke", can_fail=True)
    def apply_select(self, plugin_invoker, exec_args=[]):
        if "--discover" in exec_args:
            return

        properties_file = plugin_invoker.files["catalog"]

        try:
            with properties_file.open() as catalog:
                schema = json.load(catalog)

            reset_executor = SelectExecutor(["!*.*"])
            select_executor = SelectExecutor(self.select)
            visit(schema, reset_executor)
            visit(schema, select_executor)

            with properties_file.open("w") as catalog:
                json.dump(schema, catalog)
        except FileNotFoundError:
            logging.warn(f"Could not select stream, catalog file is missing.")
        except Exception as err:
            properties_file.unlink()
            raise PluginExecutionError(
                f"Could not select stream, catalog file is invalid: {properties_file}"
            ) from err



import os
import json
import sqlparse
import time
import glob

import networkx as nx

from copy import deepcopy
from networkx.readwrite import json_graph
from pathlib import Path
from pyhocon import ConfigFactory
from typing import Dict, List

from meltano.core.sql.design_helper import visit, PypikaJoinExecutor
from meltano.core.utils import encode_id_from_file_path
from meltano.core.utils import slugify
from meltano.core.project import Project
from meltano.core.plugin.model import Package


class MeltanoAnalysisFileParserError(Exception):
    def __init__(self, message, file_name):
        self.message = message
        self.file_name = file_name
        super().__init__(self.message)


class MeltanoAnalysisFileParserMissingTableError(MeltanoAnalysisFileParserError):
    def __init__(self, field, your_choice, cls, file_name):
        self.file_name = file_name
        self.cls = cls
        self.message = f'Missing accompanying table "{your_choice}" in "{field}" field in {cls} in {file_name}.'
        super().__init__(self.message, self.file_name)


class MeltanoAnalysisMissingTopicFilesError(MeltanoAnalysisFileParserError):
    def __init__(self, message):
        super().__init__(message, "topic file")


class MeltanoAnalysisFileParserUnacceptableChoiceError(MeltanoAnalysisFileParserError):
    def __init__(self, field, your_choice, acceptable_choices, cls, file_name):
        self.file_name = file_name
        self.cls = cls
        quoted_choices = [f'"{a}"' for a in acceptable_choices]
        self.message = f"Unacceptable choice for field {field}. You wrote \"{your_choice}\". Acceptable choices are: {', '.join(quoted_choices)}. In {file_name}"
        super().__init__(self.message, self.file_name)


class MeltanoAnalysisFileParserMissingFieldsError(MeltanoAnalysisFileParserError):
    def __init__(self, fields, cls, file_name):
        self.file_name = file_name
        self.cls = cls
        field = "field" if len(fields) == 1 else "fields"
        self.message = (
            f"Missing {field} in {cls}: \"{', '.join(fields)}\" in {self.file_name}"
        )
        super().__init__(self.message, self.file_name)

    def __str__(self):
        return f"{self.message} in {self.file_name}"

    def __repr__(self):
        return f"{self.message} in {self.file_name}"


class MeltanoAnalysisInvalidJoinDependencyError(MeltanoAnalysisFileParserError):
    def __init__(self, field, cls, file_name):
        self.file_name = file_name
        self.cls = cls
        self.message = (
            f'Invalid join dependency in "{file_name}" in "{field}" join in {cls}.'
        )
        super().__init__(self.message, self.file_name)


class MeltanoAnalysisFileParser:
    def __init__(self, project: Project):
        self.project = project
        self._output_dir = self.project.root_dir("model")
        self.topics = []
        self.packaged_topics = []
        self.required_topic_properties = ["name", "connection", "label", "designs"]
        self.required_design_properties = ["from", "label", "description"]
        self.required_join_properties = ["sql_on", "relationship"]
        self.required_table_properties = ["sql_table_name", "columns"]
        self.join_relationship_types = [
            "one_to_one",
            "one_to_many",
            "many_to_one",
            "many_to_many",
        ]
        self.project = Project()

    def uses_accepted_choice(self, choices, choice):
        return choice in choices

    def missing_properties(self, properties, properties_dict):
        properties_copy = properties.copy()
        for prop in properties_dict:
            try:
                property_index = properties_copy.index(prop)
                del properties_copy[property_index]
            except ValueError as e:
                continue
        return properties_copy

    def parse_m5o_file(self, file_path):
        try:
            return ConfigFactory.parse_file(file_path)
        except Exception as e:
            raise MeltanoAnalysisFileParserError(str(e), str(file_path.parts[-1]))

    def generate_join_graph_for_node(self, graph, node, joins):
        remaining_joins = []
        check_against = []
        for join in joins:
            join_executor = None
            if "executor" in join:
                join_executor = join["executor"]
            else:
                join_executor = PypikaJoinExecutor(self, join)
                visit(sqlparse.parse(join["sql_on"])[0], join_executor)
                join["executor"] = join_executor
            left_table = join_executor.comparison_fields.left.table
            right_table = join_executor.comparison_fields.right.table
            sql_on = join_executor.join["sql_on"]
            if left_table == node:
                graph.add_node(right_table)
                graph.add_edge(node, right_table)
                check_against.append(right_table)
            elif right_table == node:
                graph.add_node(left_table)
                graph.add_edge(node, left_table)
                check_against.append(left_table)
            else:
                remaining_joins.append(join)

        if len(check_against):
            for check in check_against:
                self.generate_join_graph_for_node(graph, check, remaining_joins)

    def graph_design(self, design):
        base_design = design["related_table"]["name"]
        graph = nx.DiGraph()
        graph.add_node(base_design)
        joins = deepcopy(design.get("joins", []))
        self.generate_join_graph_for_node(graph, base_design, joins)
        return json_graph.node_link_data(graph)

    def graph_topic(self, topic):
        designs = topic["designs"]
        for design in designs:
            design_graph = self.graph_design(design)
            design["graph"] = design_graph
        return topic

    def compile(self, topics):
        indices = {}
        for topic in topics:
            compiled_file_name = f"{topic['name']}.topic.m5oc"
            compiled_file_path = self._output_dir.joinpath(compiled_file_name)
            indices[topic["name"]] = {"designs": [e["name"] for e in topic["designs"]]}
            topic = self.graph_topic(topic)
            with compiled_file_path.open("w") as compiled_topic:
                compiled_topic.write(json.dumps(topic))

        # index file
        index_file_path = self._output_dir.joinpath("topics.index.m5oc")
        with index_file_path.open("w") as index_file:
            index_file.write(json.dumps(indices))

    def parse_packages(self):
        for package in self.packages():
            self.m5o_tables = package.tables
            self.m5o_topics = package.topics
            if not self.m5o_topics and self.m5o_tables:
                raise MeltanoAnalysisMissingTopicFilesError(
                    f"Missing topic file(s) for package {package}"
                )
            for topic in self.m5o_topics:
                file_name = topic.parts[-1]
                conf = self.parse_m5o_file(topic)
                parsed_topic = self.topic(conf, file_name)
                self.packaged_topics.append(parsed_topic)

        return self.packaged_topics

    def parse(self):
        models = self.project.root_dir("model")
        self.m5o_tables = list(models.glob("*.table.m5o"))
        self.m5o_topics = list(models.glob("*.topic.m5o"))
        if not self.m5o_topics and self.m5o_tables:
            raise MeltanoAnalysisMissingTopicFilesError("Missing topic file(s)")
        for topic in self.m5o_topics:
            file_name = topic.parts[-1]
            conf = self.parse_m5o_file(topic)
            parsed_topic = self.topic(conf, file_name)
            self.topics.append(parsed_topic)
        return self.topics

    def topic(self, ma_file_topic_dict, file_name):
        temp_topic = {}
        missing_properties = self.missing_properties(
            self.required_topic_properties, ma_file_topic_dict
        )
        if missing_properties:
            raise MeltanoAnalysisFileParserMissingFieldsError(
                missing_properties, "topic", file_name
            )
        for prop_name, prop_def in ma_file_topic_dict.items():
            temp_topic[prop_name] = prop_def
            if prop_name == "designs":
                temp_topic[prop_name] = self.designs(prop_def, file_name)
        return temp_topic

    def table_from_full_table_name(self, table_name):
        delimiter = "."
        if not delimiter in table_name:
            return table_name
        else:
            return table_name.split(".")[1]

    def table_conf_by_name(self, table_name, cls, prop, file_name):
        # grab just the table from schema table
        table_name = self.table_from_full_table_name(table_name)
        try:
            return next(
                table
                for table in self.m5o_tables
                if table.parts[-1] == f"{table_name}.table.m5o"
            )
        except StopIteration as e:
            raise MeltanoAnalysisFileParserMissingTableError(
                prop, table_name, cls, file_name
            )

    def designs(self, ma_file_designs_dict, file_name):
        topic_designs = []
        for design_name, design_def in ma_file_designs_dict.items():
            temp_design = {}
            temp_design["name"] = design_name
            missing_properties = self.missing_properties(
                self.required_design_properties, design_def
            )
            if missing_properties:
                raise MeltanoAnalysisFileParserMissingFieldsError(
                    missing_properties, "design", file_name
                )
            for prop_name, prop_def in design_def.items():
                temp_design[prop_name] = prop_def
                if prop_name == "from":
                    matching_table = self.table_conf_by_name(
                        temp_design[prop_name], "design", prop_name, file_name
                    )
                    temp_design["related_table"] = self.table(
                        self.parse_m5o_file(matching_table), matching_table.parts[-1]
                    )
                if prop_name == "joins":
                    temp_design[prop_name] = self.joins(prop_def, file_name)
            topic_designs.append(temp_design)
        return topic_designs

    def joins(self, ma_file_joins_dict, file_name):
        design_joins = []
        for join_name, join_def in ma_file_joins_dict.items():
            temp_join = {}
            temp_join["name"] = join_name
            related_table_name = join_def.get("from", join_name)
            matching_table = self.table_conf_by_name(
                related_table_name, "join", "name", file_name
            )
            temp_join["related_table"] = self.table(
                self.parse_m5o_file(matching_table), matching_table.parts[-1]
            )
            missing_properties = self.missing_properties(
                self.required_join_properties, join_def
            )
            if missing_properties:
                raise MeltanoAnalysisFileParserMissingFieldsError(
                    missing_properties, "join", file_name
                )
            for prop_name, prop_def in join_def.items():
                temp_join[prop_name] = prop_def
                if prop_name == "relationship":
                    uses_accepted_choices = self.uses_accepted_choice(
                        self.join_relationship_types, prop_def
                    )
                    if not uses_accepted_choices:
                        raise MeltanoAnalysisFileParserUnacceptableChoiceError(
                            prop_name,
                            prop_def,
                            self.join_relationship_types,
                            "join",
                            file_name,
                        )

            design_joins.append(temp_join)
        return design_joins

    def table(self, table_file, file_name):
        temp_table = {}
        missing_properties = self.missing_properties(
            self.required_table_properties, table_file
        )
        if missing_properties:
            raise MeltanoAnalysisFileParserMissingFieldsError(
                missing_properties, "table", file_name
            )

        for prop_name, prop_def in table_file.items():
            subproperties = ("columns", "aggregates", "timeframes")

            if prop_name in subproperties:
                temp_table[prop_name] = self.name_flatten_dict(prop_def)
            else:
                temp_table[prop_name] = prop_def

        return temp_table

    @classmethod
    def name_flatten_dict(cls, d: Dict) -> List:
        return [{"name": k, **rest} for k, rest in d.items()]

    @staticmethod
    def fill_base_m5o_dict(file, name, file_dict=None):
        if file_dict is None:
            file_dict = {"name": name}

        file_dict["path"] = str(file)
        file_dict["id"] = encode_id_from_file_path(file_dict["path"])
        file_dict["slug"] = slugify(name)
        file_dict["createdAt"] = time.time()
        return file_dict

    def packages(self) -> List[Package]:
        return [
            Package(f.name, self.project)
            for f in self.project.model_dir().iterdir()
            if f.is_dir()
        ]

    @classmethod
    def package_files(cls, package):
        m5oFiles = {
            "topics": {"label": "Topics", "items": []},
            "tables": {"label": "Tables", "items": []},
        }
        for f in package.files:
            basename = os.path.basename(f)
            filename, ext = os.path.splitext(basename)
            if ext != ".m5o":
                continue

            # filename splittext twice occurs due to current *.type.extension convention (two dots)
            filename = f"{package.name}/{filename.lower()}"
            filename, ext = os.path.splitext(filename)
            file_dict = cls.fill_base_m5o_dict(f, filename)
            if ext == ".topic":
                m5oFiles["topics"]["items"].append(file_dict)
            if ext == ".table":
                m5oFiles["tables"]["items"].append(file_dict)

        return m5oFiles



import os
import functools
import click
import urllib

from meltano.core.project import Project
from meltano.core.utils import pop_all


def db_options(func):
    @click.option(
        "-B",
        "--backend",
        envvar="MELTANO_BACKEND",
        default="sqlite",
        type=click.Choice(["sqlite", "postgresql"]),
        help="Database backend for Meltano.",
    )
    @click.option("--path", envvar="SQLITE_DATABASE", default="meltano")
    @click.option(
        "-H",
        "--host",
        envvar="PG_ADDRESS",
        default="localhost",
        help="Database address.",
    )
    @click.option("-p", "--port", type=int, envvar="PG_PORT", default=5432)
    @click.option(
        "-d",
        "-db",
        "database",
        envvar="PG_DATABASE",
        help="Database to import the data to.",
    )
    @click.option(
        "-u",
        "--username",
        envvar="PG_USERNAME",
        default=lambda: os.getenv("USER", ""),
        help="Specifies the user to connect to the database with.",
    )
    @click.password_option(prompt=False, envvar="PG_PASSWORD")
    def wrapper(*args, **kwargs):
        engine_uri = os.getenv("SQL_ENGINE_URI")
        backend = kwargs.pop("backend")

        config = {
            "sqlite": pop_all(("path",), kwargs),
            "postgresql": pop_all(
                ("host", "port", "database", "username", "password"), kwargs
            ),
        }

        if not engine_uri and backend == "sqlite" and config[backend]["path"]:
            engine_uri = "sqlite:///{path}.db".format(**config[backend])

        if not engine_uri and backend == "postgresql":
            pg_config = config[backend]
            pg_config["password"] = urllib.parse.quote(pg_config["password"])
            engine_uri = "postgresql://{username}:{password}@{host}:{port}/{database}".format(
                **pg_config
            )

        return func(*args, **kwargs, engine_uri=engine_uri)

    return functools.update_wrapper(wrapper, func)


def project(func):
    @click.pass_context
    def decorate(ctx, *args, **kwargs):
        project = ctx.obj["project"]
        if not project:
            raise click.ClickException(
                f"`{ctx.command_path}` must be run inside a Meltano project."
                "\nUse `meltano init <project_name>` to create one."
            )

        ctx.invoke(func, project, *args, **kwargs)

    return functools.update_wrapper(decorate, func)



import pytest

from meltano.cli import cli
from asserts import assert_cli_runner


# Use One Class per End to End test in order to setup a new project each time
class TestFullInstall:
    @pytest.mark.slow
    @pytest.mark.backend("postgresql")
    def test_carbon_intensity_postgres_dbt(request, cli_runner, monkeypatch, project):
        monkeypatch.setenv("PG_SCHEMA", "carbon")
        monkeypatch.setenv("PG_DATABASE", "pytest")
        monkeypatch.setenv("MELTANO_BACKEND", "postgresql")

        cli_args = ["add", "extractor", "tap-carbon-intensity"]
        result = cli_runner.invoke(cli, cli_args)
        assert_cli_runner(result)

        cli_args = ["add", "loader", "target-postgres"]
        result = cli_runner.invoke(cli, cli_args)
        assert_cli_runner(result)

        cli_args = ["add", "transformer", "dbt"]
        result = cli_runner.invoke(cli, cli_args)
        assert_cli_runner(result)

        cli_args = ["add", "model", "model-carbon-intensity"]
        result = cli_runner.invoke(cli, cli_args)
        assert_cli_runner(result)

        cli_args = [
            "elt",
            "tap-carbon-intensity",
            "target-postgres",
            "--transform",
            "run",
        ]
        result = cli_runner.invoke(cli, cli_args)
        assert_cli_runner(result)



import json
import os
from meltano.core.behavior.hookable import hook
from meltano.core.project import Project
from meltano.core.plugin import Plugin


class SingerPlugin(Plugin):
    def __init__(self, *args, **kwargs):
        super().__init__(self.__class__.__plugin_type__, *args, **kwargs)

    @hook("before_install")
    def install_config_stub(self, project):
        plugin_dir = project.plugin_dir(self)
        os.makedirs(plugin_dir, exist_ok=True)

        # TODO: refactor as explicit stubs
        with open(plugin_dir.joinpath(self.config_files["config"]), "w") as config:
            json.dump(self.config, config)



import aiohttp
import asyncio
import yaml

from sqlalchemy import (
    MetaData,
    Table,
    String,
    Column,
    TIMESTAMP,
    Float,
    Integer,
    Boolean,
    Date,
    REAL,
    SMALLINT,
    TEXT,
    BIGINT,
    JSON,
)
from typing import Generator, Iterator


async def fetch(url, session):
    async with session.get(url) as response:
        if response.status != 200:
            response.raise_for_status()
        return await response.text()


async def get_futures(urls: Iterator[str], timeout=None, headers=None):
    tasks = []
    async with aiohttp.ClientSession(headers=headers, timeout=timeout) as session:
        for url in urls:
            task = asyncio.ensure_future(fetch(url, session))
            tasks.append(task)
        responses = await asyncio.gather(*tasks)
    return responses


def fetch_urls(
    urls: list, headers: dict = None, timeout: aiohttp.ClientTimeout = None
) -> Iterator[str]:
    """
    Asynchronously fetch multiple urls and return Iterator object that contains all of the responses
    Example usage:
    urls = ['https://httpbin.org/get', 'https://google.com', 'https://gitlab.com/meltano/']
    for result in fetch_urls(urls):
        print(result)

    :param headers: dict of headers to be passed with the request
    :param timeout: timeout object to be passed to the aiohttp lib
    :param urls: Iterator of url strings
    :return: Generator of response texts
    """
    try:
        loop = asyncio.get_event_loop()
    except RuntimeError as e:
        # Fix for when asyncio runs inside a sub-thread so there is no
        #  event_loop available
        asyncio.set_event_loop(asyncio.new_event_loop())
        loop = asyncio.get_event_loop()

    futures = get_futures(urls=urls, headers=headers, timeout=timeout)
    future = asyncio.ensure_future(futures)
    responses = loop.run_until_complete(future)
    for resp in responses:
        yield resp


def get_sqlalchemy_col(
    column_name: str, column_type_name: str, primary_key_col_name: str
) -> Column:
    """
    Helper method that returns the sqlalchemy Column object to be used for Table def.
    """
    if column_type_name == "timestamp without time zone":
        col = Column(column_name, TIMESTAMP(timezone=False))
    elif column_type_name == "timestamp with time zone":
        col = Column(column_name, TIMESTAMP(timezone=True))
    elif column_type_name == "date":
        col = Column(column_name, Date)
    elif column_type_name == "real":
        col = Column(column_name, REAL)
    elif column_type_name == "integer":
        col = Column(column_name, Integer)
    elif column_type_name == "smallint":
        col = Column(column_name, SMALLINT)
    elif column_type_name == "text":
        col = Column(column_name, TEXT)
    elif column_type_name == "bigint":
        col = Column(column_name, BIGINT)
    elif column_type_name == "float":
        col = Column(column_name, Float)
    elif column_type_name == "boolean":
        col = Column(column_name, Boolean)
    elif column_type_name == "json":
        col = Column(column_name, JSON)
    else:
        col = Column(column_name, String)

    if column_name == primary_key_col_name:
        col.primary_key = True
    return col


def tables_from_manifest(
    manifest_file_path: str, metadata: MetaData, schema_name: str
) -> {str: Table}:
    with open(manifest_file_path) as file:
        schema_manifest: dict = yaml.load(file)
        tables = {}
        for table_name in schema_manifest:
            primary_key_col_name = schema_manifest[table_name]["primary_key"]
            columns: {str: str} = schema_manifest[table_name]["columns"]
            columns_list = [
                get_sqlalchemy_col(column_name, column_type_name, primary_key_col_name)
                for column_name, column_type_name in columns.items()
            ]

            table = Table(table_name, metadata, *columns_list, schema=schema_name)
            tables[table_name] = table
    return tables



import pytest

from meltano.core.utils import nest


def test_nest():
    subject = {}

    b = nest(subject, "a.b")
    b["val"] = 1
    assert b == {"val": 1}

    c = nest(subject, "a.b.c")
    c["val"] = 2
    assert b == {"val": 1, "c": {"val": 2}}

    arr = nest(subject, "a.list", value=[])

    VALUE = {"value": 1}
    val = nest(subject, "a.value", value=VALUE)

    assert subject["a"]["b"] is b
    assert subject["a"]["b"]["c"] is c
    assert isinstance(arr, list)
    # make sure it is a copy
    assert val == VALUE and not val is VALUE



import pytest
from unittest import mock
from contextlib import contextmanager
from flask import request_started
from flask_security.utils import login_user, logout_user

import meltano.api.app
from meltano.api.security.identity import create_dev_user
from meltano.api.models import db


def _cleanup(app):
    with app.app_context():
        db.drop_all()


@pytest.fixture
def impersonate(app):
    @contextmanager
    def factory(user):
        def push(sender):
            if user:
                login_user(user)
            else:
                logout_user()

        with request_started.connected_to(push):
            yield

    return factory


@pytest.fixture()
def app(create_app):
    return create_app()


@pytest.fixture()
def app_context(app):
    with app.app_context():
        yield


@pytest.fixture()
def create_app(request, add_model, project):
    def _factory(**kwargs):
        config = {
            "TESTING": True,
            "LOGIN_DISABLED": False,
            "ENV": "test",
            "SQLALCHEMY_DATABASE_URI": "sqlite://",
            **kwargs,
        }  # in-memory

        app = meltano.api.app.create_app(config)
        request.addfinalizer(lambda: _cleanup(app))

        with app.app_context():
            db.drop_all()
            db.create_all()
            create_dev_user()

        return app

    return _factory


@pytest.fixture
def api(app):
    return app.test_client()



import json
import pytest
from flask import url_for


def assert_has_items(entry, count):
    return len(entry["items"]) == count


def test_index(api, app):
    with app.test_request_context():
        res = api.get(url_for("repos.index"))

    payload = res.json

    assert_has_items(payload["tables"], 3)
    assert_has_items(payload["topics"], 1)
    assert_has_items(payload["dashboards"], 0)
    assert_has_items(payload["documents"], 1)


def test_models(api, app):
    with app.test_request_context():
        res = api.get(url_for("repos.models"))

    payload = res.json

    # we have topic
    assert payload.keys()

    # each topics has designs
    for topic_def in payload.values():
        assert topic_def["designs"]


def test_design_read(api, app):
    with app.test_request_context():
        res = api.get(
            url_for("repos.design_read", topic_name="carbon", design_name="region")
        )

    json_data = json.loads(res.data)

    assert "description" in json_data
    assert "from" in json_data
    assert "graph" in json_data
    assert "joins" in json_data
    assert "label" in json_data
    assert "name" in json_data
    assert "related_table" in json_data

    assert json_data["from"] == "region"
    assert json_data["name"] == "region"



import pytest
import re

from flask import url_for


@pytest.fixture()
def compile_models(api):
    api.get("/repos/sync")


def assertIsSQL(value: str) -> bool:
    assert (
        re.match(r"SELECT.*FROM.*(JOIN.*)*(GROUP BY.*)?(LIMIT \d+)?", value),
        f"{value} is not a SQL query.",
    )


@pytest.mark.usefixtures("project", "add_model")
class TestSqlController:
    @pytest.fixture
    def post(self, app, api):
        def _post(payload):
            return api.post(self.url(app, "carbon", "region"), json=payload)

        return _post

    @classmethod
    def url(cls, app, topic, design):
        with app.test_request_context():
            return url_for("sql.get_sql", topic_name=topic, design_name=design)

    def test_get_sql(self, post):
        self.assert_empty_query(post)
        self.assert_column_query(post)
        self.assert_aggregate_query(post)
        self.assert_timeframe_query(post)
        self.assert_join_graph_no_dependencies(post)
        self.assert_join_graph_one_dependency(post)
        self.assert_join_graph_two_dependency(post)
        self.assert_join_graph_derived_dependency(post)
        self.assert_no_join(post)

    def assert_empty_query(self, post):
        """with no columns no query should be generated"""

        payload = {
            "table": "region",
            "columns": [],
            "aggregates": [],
            "timeframes": [],
            "joins": [
                {"name": "entry", "columns": []},
                {"name": "generationmix", "columns": []},
            ],
            "order": None,
            "limit": 3,
            "filters": {},
            "run": False,
        }

        res = post(payload)
        assert res.status_code == 200, res.data
        assert res.json["sql"] == ""

    def assert_column_query(self, post):
        """with columns they should be included in the query"""

        payload = {
            "table": "region",
            "columns": ["name"],
            "aggregates": [],
            "timeframes": [],
            "joins": [
                {"name": "entry", "columns": ["forecast"]},
                {"name": "generationmix", "columns": ["perc", "fuel"]},
            ],
            "order": None,
            "limit": 3,
            "filters": {},
            "run": False,
        }

        res = post(payload)

        assert res.status_code == 200
        assertIsSQL(res.json["sql"])
        assert '"region.dnoregion"' in res.json["sql"]
        assert '"entry.forecast"' in res.json["sql"]
        assert '"generationmix.perc"' in res.json["sql"]
        assert '"generationmix.fuel"' in res.json["sql"]

    def assert_aggregate_query(self, post):
        """with aggregates they should be included in the query"""

        payload = {
            "table": "region",
            "columns": ["name"],
            "aggregates": ["count"],
            "timeframes": [],
            "joins": [
                {"name": "entry", "columns": [], "aggregates": [], "timeframes": []},
                {"name": "generationmix", "columns": [], "aggregates": []},
            ],
            "order": None,
            "limit": 3,
            "filters": {},
            "run": False,
        }

        res = post(payload)

        assert res.status_code == 200
        assertIsSQL(res.json["sql"])
        assert "region.count" in res.json["aggregates"]
        assert (
            '"region.dnoregion",COALESCE(COUNT("region"."id"),0) "region.count" FROM "region" "region"'
            in res.json["sql"]
        )

    def assert_timeframe_query(self, post):
        payload = {
            "table": "region",
            "columns": ["name"],
            "aggregates": [],
            "timeframes": [],
            "joins": [
                {
                    "name": "entry",
                    "columns": ["forecast"],
                    "timeframes": [
                        {
                            "name": "from",
                            "periods": [{"label": "Week", "selected": True}],
                        },
                        {"name": "to", "periods": []},
                    ],
                },
                {"name": "generationmix", "columns": ["perc", "fuel"]},
            ],
            "order": None,
            "limit": 3,
            "filters": {},
            "run": False,
        }

        res = post(payload)
        assert res.status_code == 200, res.data
        assertIsSQL(res.json["sql"])
        assert 'EXTRACT(\'Week\' FROM "entry"."from") "from.week"' in res.json["sql"]
        assert re.search(r'GROUP BY.*"from.week"', res.json["sql"])

    def assert_join_graph_no_dependencies(self, post):
        payload = {
            "table": "region",
            "columns": ["name"],
            "aggregates": [],
            "timeframes": [],
            "joins": [],
            "order": None,
            "limit": 3,
            "filters": {},
            "run": False,
        }

        res = post(payload)

        assert res.status_code == 200
        assert (
            'SELECT "region"."dnoregion" "region.dnoregion" FROM "region" "region" GROUP BY "region.dnoregion" LIMIT 3;'
            in res.json["sql"]
        )

    def assert_join_graph_one_dependency(self, post):
        payload = {
            "table": "region",
            "columns": ["name"],
            "aggregates": [],
            "timeframes": [],
            "joins": [{"name": "entry", "columns": ["forecast"]}],
            "order": None,
            "limit": 3,
            "filters": {},
            "run": False,
        }

        res = post(payload)

        assert res.status_code == 200
        sql = 'SELECT "region"."dnoregion" "region.dnoregion","entry"."forecast" "entry.forecast" FROM "region" "region" JOIN "entry" "entry" ON "region"."id"="entry"."region_id" GROUP BY "region.dnoregion","entry.forecast" LIMIT 3;'
        assert sql in res.json["sql"]

    def assert_join_graph_two_dependency(self, post):
        payload = {
            "table": "region",
            "columns": ["name"],
            "aggregates": [],
            "timeframes": [],
            "joins": [
                {"name": "entry", "columns": ["forecast"]},
                {"name": "generationmix", "columns": ["perc", "fuel"]},
            ],
            "order": None,
            "limit": 3,
            "filters": {},
            "run": False,
        }

        res = post(payload)

        assert res.status_code == 200
        sql = 'SELECT "region"."dnoregion" "region.dnoregion","entry"."forecast" "entry.forecast","generationmix"."fuel" "generationmix.fuel","generationmix"."perc" "generationmix.perc" FROM "region" "region" JOIN "entry" "entry" ON "region"."id"="entry"."region_id" JOIN "generationmix" "generationmix" ON "entry"."id"="generationmix"."entry_id" GROUP BY "region.dnoregion","entry.forecast","generationmix.fuel","generationmix.perc" LIMIT 3;'
        assert sql in res.json["sql"]

    def assert_join_graph_derived_dependency(self, post):
        payload = {
            "table": "region",
            "columns": ["name"],
            "aggregates": [],
            "timeframes": [],
            "joins": [
                {"name": "entry", "columns": []},
                {"name": "generationmix", "columns": ["perc", "fuel"]},
            ],
            "order": None,
            "limit": 3,
            "filters": {},
            "run": False,
        }

        res = post(payload)

        assert res.status_code == 200
        sql = 'SELECT "region"."dnoregion" "region.dnoregion","generationmix"."fuel" "generationmix.fuel","generationmix"."perc" "generationmix.perc" FROM "region" "region" JOIN "entry" "entry" ON "region"."id"="entry"."region_id" JOIN "generationmix" "generationmix" ON "entry"."id"="generationmix"."entry_id" GROUP BY "region.dnoregion","generationmix.fuel","generationmix.perc" LIMIT 3;'
        assert sql in res.json["sql"]

    def assert_no_join(self, post):
        payload = {
            "table": "region",
            "columns": ["name"],
            "aggregates": [],
            "timeframes": [],
            "joins": [],
            "order": None,
            "limit": 3,
            "filters": {},
            "run": False,
        }

        res = post(payload)

        assert res.status_code == 200
        sql = 'SELECT "region"."dnoregion" "region.dnoregion" FROM "region" "region" GROUP BY "region.dnoregion" LIMIT 3;'
        assert sql in res.json["sql"]



import json
import os

from os.path import join
from pathlib import Path

from meltano.core.m5o.m5oc_file import M5ocFile
from meltano.core.m5o.m5o_file_parser import MeltanoAnalysisFileParser
from meltano.core.m5o.m5o_collection_parser import (
    M5oCollectionParser,
    M5oCollectionParserTypes,
)
from meltano.core.project import Project
from meltano.core.utils import slugify
from meltano.core.m5o.m5o_collection_parser import (
    M5oCollectionParser,
    M5oCollectionParserTypes,
)


class ReportsHelper:
    VERSION = "1.0.0"

    def has_reports(self):
        project = Project.find()
        m5oc_file = project.root_dir("model", "reports.m5oc")
        return Path.is_file(m5oc_file)

    def get_report_m5oc(self):
        project = Project.find()
        m5oc_file = project.root_dir("model", "reports.m5oc")
        return M5ocFile.load(m5oc_file)

    def get_reports(self):
        project = Project.find()
        path = project.root_dir("model")
        reportsParser = M5oCollectionParser(path, M5oCollectionParserTypes.Report)
        return reportsParser.contents()

    def load_report(self, report_name):
        reports = self.get_reports()
        target_report = [report for report in reports if report["name"] == report_name]
        return target_report[0]

    def save_report(self, data):
        project = Project.find()
        slug = slugify(data["name"])
        file_name = f"{slug}.report.m5o"
        file_path = project.root_dir("model", file_name)
        data = MeltanoAnalysisFileParser.fill_base_m5o_dict(file_path, slug, data)
        data["version"] = ReportsHelper.VERSION
        with open(file_path, "w") as f:
            json.dump(data, f)
        return data

    def update_report(self, data):
        project = Project.find()
        file_name = f"{data['slug']}.report.m5o"
        file_path = project.root_dir("model", file_name)
        with open(file_path, "w") as f:
            json.dump(data, f)
        return data



import logging
import os
import sqlalchemy
from pathlib import Path
from collections import OrderedDict
from flask import jsonify, redirect, url_for
from pypika import Query, Order

from meltano.api.models import db
from meltano.api.security import create_dev_user
from meltano.core.project import Project
from meltano.core.m5o.m5oc_file import M5ocFile
from meltano.core.sql.analysis_helper import AnalysisHelper
from meltano.core.sql.sql_utils import SqlUtils
from .settings_helper import SettingsHelper


class ConnectionNotFound(Exception):
    def __init__(self, connection_name: str):
        self.connection_name = connection_name
        super().__init__("{connection_name} is missing.")


class UnsupportedConnectionDialect(Exception):
    def __init__(self, connection_dialect: str):
        self.connection_dialect = connection_dialect
        super().__init__("Dialect {connection_dialect} is not supprted.")


class SqlHelper(SqlUtils):
    def parse_sql(self, input):
        placeholders = self.placeholder_match(input)

    def placeholder_match(self, input):
        outer_pattern = r"(\$\{[\w\.]*\})"
        inner_pattern = r"\$\{([\w\.]*)\}"
        outer_results = re.findall(outer_pattern, input)
        inner_results = re.findall(inner_pattern, input)
        return (outer_results, inner_results)

    def get_m5oc_topic(self, topic_name):
        project = Project.find()
        m5oc_file = project.root_dir("model", f"{topic_name}.topic.m5oc")
        return M5ocFile.load(m5oc_file)

    def get_connection(self, connection_name):
        settings_helper = SettingsHelper()
        connections = settings_helper.get_connections()["settings"]["connections"]

        try:
            connection = next(
                connection
                for connection in connections
                if connection["name"] == connection_name
            )

            return connection
        except StopIteration:
            raise ConnectionNotFound(connection_name)

    def get_db_engine(self, connection_name):
        project = Project.find()
        connection = self.get_connection(connection_name)

        if connection["dialect"] == "postgresql":
            psql_params = ["username", "password", "host", "port", "database"]
            user, pw, host, port, db = [connection[param] for param in psql_params]
            connection_url = f"postgresql+psycopg2://{user}:{pw}@{host}:{port}/{db}"
        elif connection["dialect"] == "sqlite":
            db_path = project.root.joinpath(connection["path"])
            connection_url = f"sqlite:///{db_path}"
        else:
            raise UnsupportedConnectionDialect(connection["dialect"])

        return sqlalchemy.create_engine(connection_url)

    def get_query_results(self, connection_name, sql):
        engine = self.get_db_engine(connection_name)
        results = engine.execute(sql)
        results = [OrderedDict(row) for row in results]
        return results

    def reset_db(self):
        try:
            db.drop_all()
        except sqlalchemy.exc.OperationalError as err:
            logging.error("Failed drop database.")

        db.create_all()
        create_dev_user()



import yaml
import fnmatch
from typing import Dict
from collections import namedtuple
from enum import Enum

from meltano.core.behavior.hookable import HookObject


class YAMLEnum(str, Enum):
    def __str__(self):
        return self.value

    @staticmethod
    def yaml_representer(dumper, obj):
        return dumper.represent_scalar("tag:yaml.org,2002:str", str(obj))


yaml.add_multi_representer(YAMLEnum, YAMLEnum.yaml_representer)


class PluginType(YAMLEnum):
    EXTRACTORS = "extractors"
    LOADERS = "loaders"
    MODELS = "models"
    TRANSFORMERS = "transformers"
    TRANSFORMS = "transforms"
    ORCHESTRATORS = "orchestrators"
    ALL = "all"

    def __str__(self):
        return self.value

    @property
    def cli_command(self):
        """Makes it singular for `meltano add PLUGIN_TYPE`"""
        if self is self.__class__.ALL:
            raise NotImplemented()

        return self.value[:-1]

    @classmethod
    def value_exists(cls, value):
        return value in cls._value2member_map_


class Plugin(HookObject):
    """
    Args:
    name: The unique name for the installed plugin
    pip_url: The pip-compatible installation URI, like `git+https://…` or `-e /path/to/pkg`
    executable: The plugin executable name (default: <name>)
    """

    def __init__(
        self,
        plugin_type: PluginType,
        name: str,
        pip_url=None,
        config=None,
        select=None,
        **extras,
    ):
        self.name = name
        self.type = plugin_type
        self.pip_url = pip_url
        self.config = config
        self._select = set(select or [])
        self._extras = extras or {}

    def canonical(self):
        canonical = {
            "name": self.name,
            "pip_url": self.pip_url,
            "config": self.config,
            **self._extras,
        }

        if self._select:
            canonical.update({"select": list(self._select)})

        if self._extras:
            canonical.update(**self._extras)

        return canonical

    def invoker(self, project, *args, **kwargs):
        "Override to have a specialize PluginInvoker class"
        pass

    def exec_args(self, files: Dict):
        return []

    @property
    def config_files(self):
        """Return a list of stubbed files created for this plugin."""
        return dict()

    @property
    def output_files(self):
        return dict()

    @property
    def select(self):
        return self._select or {"*.*"}

    @select.setter
    def select(self, patterns):
        self._select = set(patterns)

    @property
    def executable(self):
        return self._extras.get("executable", self.name)

    def cwd(self, project):
        return project.root

    def add_select_filter(self, filter: str):
        self._select.add(filter)

    def __eq__(self, other):
        return self.name == other.name and self.type == other.type

    def __repr__(self):
        return f"<Plugin name={self.name} type={self.type}>"



import pytest

from meltano.cli import cli
from asserts import assert_cli_runner


class TestAutoInstall:
    @pytest.mark.slow
    @pytest.mark.skip(reason="Deadlock on docker, have to investigate further.")
    def test_carbon_intensity_sqlite(request, cli_runner, monkeypatch, project):
        # make the job log in-memory
        monkeypatch.setenv("SQL_ENGINE_URI", "sqlite://")

        # Directly run `meltano add loader target-sqlite` in a new project
        # So that the extractor and the loader are auto installed by meltano
        cli_args = ["elt", "tap-carbon-intensity", "target-sqlite"]
        result = cli_runner.invoke(cli, cli_args)

        assert_cli_runner(result)



import pytest
import os
import shutil
from copy import copy

import meltano
from meltano.cli import cli


@pytest.fixture
def project(test_dir, project_init_service):
    """This fixture returns the non-activated project."""
    project = project_init_service.init()

    yield project

    os.chdir(test_dir)
    shutil.rmtree(project.root)


def test_activate_project(project, cli_runner, pushd):
    # `cd` into a project
    pushd(project.root)

    # let's overwrite the `.env` to add a sentinel value
    with project.root.joinpath(".env").open("w") as env:
        env.write("CLI_TEST_ACTIVATE_PROJECT=1")

    # run any cli command - that should activate the project
    cli_runner.invoke(cli, ["install"])

    assert os.getenv("CLI_TEST_ACTIVATE_PROJECT") == "1"


def test_version(cli_runner):
    cli_version = cli_runner.invoke(cli, ["--version"])

    assert cli_version.output == f"meltano, version {meltano.__version__}\n"



# DEPRECATED: do we have any good reason to keep this
# method of booting the web ui?
# We now have the `meltano ui` command that
# should take care of this.
from meltano.api.app import start
from meltano.core.project import Project


start(Project.find(), host="0.0.0.0")



import os
import subprocess
import click

from . import cli
from .params import project
from meltano.api.app import start
from meltano.core.tracking import GoogleAnalyticsTracker


@cli.command()
@project
@click.option("--debug/--no-debug", default=True, help="Run webserver in debug mode")
@click.option(
    "--port",
    default=5000,
    help="Port to run webserver on",
    envvar="MELTANO_API_PORT",
    type=int,
)
@click.option(
    "--reload/--no-reload",
    default=True,
    help="To reload the server or not on file changes",
)
@click.option(
    "--hostname",
    default="0.0.0.0",
    help="The hostname (or IP address) to bind on",
    envvar="MELTANO_API_HOSTNAME",
)
def ui(project, debug, port, reload, hostname):
    tracker = GoogleAnalyticsTracker(project)
    tracker.track_meltano_ui()

    # todo: run gunicorn if not in debug mode
    start(project, debug=debug, use_reloader=reload, port=port, host=hostname)



from meltano.core.plugin_discovery_service import PluginDiscoveryService
from meltano.core.plugin import PluginType
from meltano.core.project import Project
from meltano.core.project_add_service import ProjectAddService
from meltano.core.plugin_install_service import PluginInstallService
from meltano.core.select_service import SelectService
from meltano.core.tracking import GoogleAnalyticsTracker

from meltano.cli.add import extractor

from flask import Blueprint, request, url_for, jsonify, make_response, Response

orchestrationsBP = Blueprint(
    "orchestrations", __name__, url_prefix="/api/v1/orchestrations"
)


@orchestrationsBP.route("/", methods=["GET"])
def index():
    new_project = Project()
    new_plugin_discovery_service = PluginDiscoveryService(new_project)
    result = new_plugin_discovery_service.discover(PluginType.ALL)
    return jsonify(result)


@orchestrationsBP.route("/installed-plugins", methods=["GET"])
def installed_plugins():
    project = Project.find()
    return jsonify(project.meltano)


@orchestrationsBP.route("/add-extractor", methods=["POST"])
def add_extractor():
    project = Project.find()
    add_service = ProjectAddService(project)
    plugin_name = request.get_json()["name"]
    plugin = add_service.add("extractors", plugin_name)
    install_service = PluginInstallService(project)
    run_venv = install_service.create_venv(plugin)
    run_install_plugin = install_service.install_plugin(plugin)
    tracker = GoogleAnalyticsTracker(project)
    tracker.track_meltano_add(plugin_type="extractor", plugin_name=plugin_name)
    return jsonify({"test": 123})


@orchestrationsBP.route("/add-loader", methods=["POST"])
def add_loader():
    project = Project.find()
    add_service = ProjectAddService(project)
    plugin_name = request.get_json()["name"]
    plugin = add_service.add("loaders", plugin_name)
    install_service = PluginInstallService(project)
    run_venv = install_service.create_venv(plugin)
    run_install_plugin = install_service.install_plugin(plugin)
    tracker = GoogleAnalyticsTracker(project)
    tracker.track_meltano_add(plugin_type="loader", plugin_name=plugin_name)
    return jsonify({"test": 123})


@orchestrationsBP.route("/connection_names", methods=["GET"])
def connection_names():
    settings = Settings.query.first()
    if not settings:
        settings = Settings()
    connections = [
        c["name"] for c in settings.serializable()["settings"]["connections"]
    ]
    return jsonify(connections)


@orchestrationsBP.route("/run", methods=["POST"])
def run():
    out_nl = "\n"
    run_output = StringIO(newline=out_nl)
    incoming = request.get_json()
    extractor_name = incoming["extractor"]
    loader_name = incoming["loader"]
    connection_name = incoming.get("connection_name")
    runner = MeltanoRunner()

    # Use a logging handler to capture the output of the extraction run
    runner.logger.addHandler(logging.StreamHandler(run_output))
    runner.perform(extractor_name, loader_name)

    run_output.write("\nLoad done!")
    run_output.write("\nStarting Transform")

    # use the extractor's name as the model name for the transform operation
    transform_log = run_transform(extractor_name, connection_name)

    if transform_log["status"] != "ok":
        run_output.write("\nERROR")
    else:
        run_output.write(f"\nStatus: {transform_log['status']}")
        run_output.write(f"\nCommand: {transform_log['command']}")

    run_output.write(f"\n\nStatus:{transform_log['output']}\n")
    return jsonify({"append": run_output.getvalue()})


@orchestrationsBP.route("/extract/<extractor_name>", methods=["POST"])
def extract(extractor_name: str) -> Response:
    """
    endpoint that performs extraction of the selected datasource to the .csv file(s)
    stored on disk in the /static/tmp/ directory)
    """
    runner = MeltanoRunner()
    csv_files = runner.perform(extractor_name, "csv")
    csv_files_url = [
        url_for("static", filename=f"tmp/{file_name}") for file_name in csv_files
    ]
    return jsonify(
        {"extractor_name": extractor_name, "output_file_paths": csv_files_url}
    )


@orchestrationsBP.route("/get/configuration", methods=["POST"])
def get_plugin_configuration() -> Response:
    """
    endpoint for getting a plugin's configuration
    """
    project = Project.find()
    incoming = request.get_json()
    plugin_name = incoming["name"]
    plugin_type = incoming["type"]
    discovery_service = PluginDiscoveryService(project)
    # TODO update when save_plugin_configuration is implemented and default to the discovery otherwise
    plugin = discovery_service.find_plugin(plugin_type, plugin_name)
    return jsonify(plugin.config)


@orchestrationsBP.route("/save/configuration", methods=["POST"])
def save_plugin_configuration() -> Response:
    """
    endpoint for persisting a plugin configuration
    """
    incoming = request.get_json()
    plugin_name = incoming["name"]
    plugin_type = incoming["type"]
    config = incoming["config"]

    # TODO persist strategy

    return jsonify({"test": True})


@orchestrationsBP.route("/select-entities", methods=["POST"])
def selectEntities() -> Response:
    """
    endpoint that performs selection of the user selected entities and attributes
    """
    project = Project.find()
    incoming = request.get_json()
    extractor_name = incoming["extractorName"]
    entity_groups = incoming["entityGroups"]

    select_service = SelectService(project, extractor_name)

    for entity_group in entity_groups:
        group_is_selected = "selected" in entity_group

        for attribute in entity_group["attributes"]:
            if group_is_selected or "selected" in attribute:
                entities_filter = entity_group["name"]
                attributes_filter = attribute["name"]
                select_service.select(entities_filter, attributes_filter)

    return jsonify("winning")


@orchestrationsBP.route("/entities/<extractor_name>", methods=["POST"])
def entities(extractor_name: str) -> Response:
    """
    endpoint that returns the entities associated with a particular extractor
    """
    project = Project.find()
    select_service = SelectService(project, extractor_name)
    list_all = select_service.get_extractor_entities()

    entity_groups = []
    for stream, prop in (
        (stream, prop)
        for stream in list_all.streams
        for prop in list_all.properties[stream.key]
    ):
        match = next(
            (
                entityGroup
                for entityGroup in entity_groups
                if entityGroup["name"] == stream.key
            ),
            None,
        )
        if match:
            match["attributes"].append({"name": prop.key})
        else:
            entity_groups.append(
                {"name": stream.key, "attributes": [{"name": prop.key}]}
            )

    entity_groups = sorted(entity_groups, key=lambda k: k["name"])
    for entityGroup in entity_groups:
        entityGroup["attributes"] = sorted(
            entityGroup["attributes"], key=lambda k: k["name"]
        )

    return jsonify({"extractor_name": extractor_name, "entity_groups": entity_groups})


@orchestrationsBP.route("/load/<loader_name>", methods=["POST"])
def load(loader_name: str) -> Response:
    """
    endpoint that performs load of the .csv file(s)(stored on disk in the /static/tmp/ directory)
    of the selected extractor  into the selected loader
    :param loader_name: the name of the loader to use in this request
    """
    incoming = request.get_json()
    try:
        extractor_name = incoming["extractor"]
        extractor_class = MeltanoRunner.import_cls(EXTRACTOR_REGISTRY[extractor_name])
        extractor = extractor_class()
    except KeyError:
        return make_response(
            jsonify(
                {
                    "response": "extractor not found",
                    "available_extractors": list(EXTRACTOR_REGISTRY.keys()),
                }
            ),
            404,
        )
    try:
        loader_class = MeltanoRunner.import_cls(LOADER_REGISTRY[loader_name])
    except KeyError:
        return make_response(
            jsonify(
                {
                    "response": "Loader not found",
                    "available_loaders": list(LOADER_REGISTRY.keys()),
                }
            ),
            404,
        )

    extractor_temp_files = [
        filename
        for filename in os.listdir(TEMP_FOLDER)
        if filename.startswith(extractor_name)
    ]
    for file_path in extractor_temp_files:
        filename = os.path.splitext(file_path)[0]
        entity_name = filename.split("--")[1]
        loader = loader_class(entity_name=entity_name, extractor=extractor)
        loader.schema_apply()
        datetime_cols = [
            col.name
            for col in loader.table.columns
            if isinstance(col.type, DATETIME_TYPES_TO_PARSE)
        ]
        df = pd.read_csv(
            os.path.join(TEMP_FOLDER, file_path), parse_dates=datetime_cols
        )
        df.replace({pd.NaT: None}, inplace=True)
        loader.load(df=df)
    return jsonify({"response": "done", "inserted_files": extractor_temp_files})


@orchestrationsBP.route("/transform/<topic_name>", methods=["POST"])
def transform(model_name):
    """
    {
        'connection_name': 'prod_dw',
    }
    Looks up the credential by the name of the datastore passed to the api

    Sets the environment variables and runs dbt in a subprocess for the given model
    dbt run --profiles-dir profile --target meltano_analyze --models <model_name>
    """
    incoming = request.get_json()
    connection_name = incoming.get("connection_name")

    return jsonify(run_transform(topic_name, connection_name))


def run_transform(topic_name, connection_name):
    settings = Settings.query.first()
    settings_connections = settings.settings["connections"]
    connection = next(
        (item for item in settings_connections if item["name"] == connection_name), None
    )

    if not connection:
        return {
            "status": "error",
            "output": f"Connection {connection_name} not found",
            "command": "",
        }

    new_env = os.environ.copy()

    if connection["dialect"] == "postgresql":
        new_env["PG_ADDRESS"] = connection["host"]
        new_env["PG_PORT"] = str(connection["port"])
        new_env["PG_USERNAME"] = connection["username"]
        new_env["PG_PASSWORD"] = str(connection["password"])
        new_env["PG_DATABASE"] = connection["database"]
        new_env["PG_SCHEMA"] = connection["schema"]

    run_command = [
        "dbt",
        "run",
        "--profiles-dir",
        f"{PROFILES_DIR}",
        "--target",
        "meltano_analyze",
    ]
    if model_name:
        run_command.extend(["--models", f"{topic_name}"])

    work_dir = os.getcwd()
    os.chdir(os.path.join(PROJECT_ROOT_DIR, TRANSFORM_DIR))

    command = " ".join(run_command)
    proc = subprocess.Popen(command, env=new_env, shell=True, stdout=subprocess.PIPE)
    transform_log = proc.stdout.read()

    os.chdir(work_dir)

    return {
        "command": str(command),
        "output": transform_log.decode("utf-8"),
        "status": "ok",
    }



import json
import logging

from .project import Project
from meltano.core.config_service import ConfigService
from meltano.core.plugin import PluginType
from meltano.core.plugin_invoker import invoker_factory
from meltano.core.plugin.singer.catalog import visit, ListSelectedExecutor


class SelectService:
    def __init__(self, project: Project, extractor: str):
        self.project = project
        self.config = ConfigService(project)
        self.extractor = self.config.get_plugin(extractor, PluginType.EXTRACTORS)

    def get_extractor(self):
        return self.extractor

    def load_schema(self):
        invoker = invoker_factory(self.project, self.extractor)
        if not invoker.files["catalog"].exists():
            logging.info("Catalog not found, trying to run the tap with --discover.")
            self.extractor.run_discovery(invoker)

        self.extractor.apply_select(invoker)
        with invoker.files["catalog"].open() as catalog:
            return json.load(catalog)

    def get_extractor_entities(self):
        invoker = invoker_factory(self.project, self.extractor)

    def get_extractor_entities(self):
        list_all = ListSelectedExecutor()

        try:
            schema = self.load_schema()
            visit(schema, list_all)
        except FileNotFoundError as e:
            logging.error(
                "Cannot find catalog: make sure the tap runs correctly with --discover; `meltano invoke TAP --discover`"
            )
            raise e

        return list_all

    def select(self, entities_filter, attributes_filter, exclude=False):
        exclude = "!" if exclude else ""
        pattern = f"{exclude}{entities_filter}.{attributes_filter}"

        with self.project.meltano_update() as meltano:
            self.extractor.add_select_filter(pattern)

            idx = next(
                i
                for i, it in enumerate(self.config.get_extractors())
                if it == self.extractor
            )
            meltano["plugins"]["extractors"][idx] = self.extractor.canonical()



import pytest
import yaml
import os
import shutil

from meltano.core.project import Project, ProjectNotFound


class TestProject:
    def test_find(self, project, mkdtemp):
        # defaults to the cwd
        found = Project.find()
        assert found == project

        # or you can specify a path
        found = Project.find(project.root)
        assert found == project

        # but it doens't recurse up, you have to be
        # at the meltano.yml level
        with pytest.raises(ProjectNotFound):
            Project.find(project.root.joinpath("model"))

        # and it fails if there isn't a meltano.yml
        with pytest.raises(ProjectNotFound):
            try:
                empty_dir = mkdtemp("meltano_empty_project")
                Project.find(empty_dir)
            finally:
                shutil.rmtree(empty_dir)

    def test_activate(self, project):
        assert os.getenv("MELTANO_PROJECT") is None

        with open(".env", "w") as env:
            env.write(f"MELTANO_PROJECT={project.root}")

        project.activate()
        assert os.getenv("MELTANO_PROJECT") == str(project.root)



import click
import logging
import warnings

# disable the psycopg2 warning
# this needs to run before `psycopg2` is imported
warnings.filterwarnings("ignore", category=UserWarning, module="psycopg2")

import meltano
from meltano.core.project import Project, ProjectNotFound
from meltano.core.utils import setup_logging

LEVELS = {
    "debug": logging.DEBUG,
    "info": logging.INFO,
    "warning": logging.WARNING,
    "error": logging.ERROR,
    "critical": logging.CRITICAL,
}


@click.group(invoke_without_command=True)
@click.option("--log-level", type=click.Choice(LEVELS.keys()), default="info")
@click.option("-v", "--verbose", count=True)
@click.version_option(version=meltano.__version__, prog_name="meltano")
@click.pass_context
def cli(ctx, log_level, verbose):
    setup_logging(log_level=LEVELS[log_level])
    ctx.ensure_object(dict)
    ctx.obj["verbosity"] = verbose

    try:
        ctx.obj["project"] = Project.find()
    except ProjectNotFound as err:
        ctx.obj["project"] = None



import pytest
import json
import logging
from unittest import mock
from pathlib import Path
from copy import deepcopy

from meltano.core.plugin import PluginType
from meltano.core.plugin.error import PluginExecutionError
from meltano.core.plugin_invoker import PluginInvoker
from meltano.core.plugin.singer.catalog import (
    visit,
    SelectExecutor,
    ListExecutor,
    ListSelectedExecutor,
    path_property,
)


LEGACY_CATALOG = """
{
  "streams": [
    {
      "tap_stream_id": "Entity",
      "stream": "entities",
      "key_properties": [
        "id"
      ],
      "schema": {
        "type": "object",
        "additionalProperties": false,
        "properties": {
          "id": {
            "type": "number"
          },
          "code": {
            "type": [
              "string",
              "null"
            ]
          },
          "name": {
            "type": [
              "string",
              "null"
            ]
          },
          "created_at": {
            "type": [
              "string",
              "null"
            ],
            "format": "date-time"
          },
          "active": {
            "type": [
              "boolean",
              "null"
            ]
          },
          "balance": {
            "type": [
              "number",
              "null"
            ]
          }
        }
      },
      "metadata": [
        {
          "breadcrumb": [
            "properties",
            "id"
          ],
          "metadata": {
              "inclusion": "automatic"
          }
        },
        {
          "breadcrumb": [
            "properties",
            "code"
          ],
          "metadata": {
            "inclusion": "available"
          }
        },
        {
          "breadcrumb": [
            "properties",
            "name"
          ],
          "metadata": {
            "inclusion": "available"
          }
        },
        {
          "breadcrumb": [
            "properties",
            "created_at"
          ],
          "metadata": {
            "inclusion": "automatic"
          }
        },
        {
          "breadcrumb": [
            "properties",
            "active"
          ],
          "metadata": {
            "inclusion": "available"
          }
        },
        {
          "breadcrumb": [
            "properties",
            "balance"
          ],
          "metadata": {
            "inclusion": "available"
          }
        }
      ],
      "replication_key": "created_at",
      "replication_method": "INCREMENTAL"
    }
  ]
}
"""

CATALOG = """
{
  "streams": [
    {
      "tap_stream_id": "Entity",
      "stream": "entities",
      "schema": {
        "type": "object",
        "additionalProperties": false,
        "properties": {
          "id": {
            "type": "number"
          },
          "code": {
            "type": [
              "string",
              "null"
            ]
          },
          "name": {
            "type": [
              "string",
              "null"
            ]
          },
          "created_at": {
            "type": [
              "string",
              "null"
            ],
            "format": "date-time"
          },
          "active": {
            "type": [
              "boolean",
              "null"
            ]
          },
          "balance": {
            "type": [
              "number",
              "null"
            ]
          }
        }
      },
      "metadata": [
        {
          "breadcrumb": [
            "properties",
            "id"
          ],
          "metadata": {
            "inclusion": "automatic"
          }
        },
        {
          "breadcrumb": [
            "properties",
            "code"
          ],
          "metadata": {
            "inclusion": "available"
          }
        },
        {
          "breadcrumb": [
            "properties",
            "name"
          ],
          "metadata": {
            "inclusion": "available"
          }
        },
        {
          "breadcrumb": [
            "properties",
            "created_at"
          ],
          "metadata": {
            "inclusion": "automatic"
          }
        },
        {
          "breadcrumb": [
            "properties",
            "active"
          ],
          "metadata": {
            "inclusion": "available"
          }
        },
        {
          "breadcrumb": [
            "properties",
            "balance"
          ],
          "metadata": {
            "inclusion": "available"
          }
        },
        {
          "breadcrumb": [],
          "metadata": {
            "table-key-properties": ["id"],
            "valid-replication-keys": ["created_at"],
            "forced-replication-method": "INCREMENTAL"
          }
        }
      ]
    }
  ]
}
"""

JSON_SCHEMA = """
{
  "streams": [
    {
      "tap_stream_id": "Entity",
      "stream": "entities",
      "schema": {
        "type": "object",
        "additionalProperties": false,
        "properties": {
          "id": {
            "type": "number"
          },
          "code": {
            "type": [
              "string",
              "null"
            ]
          },
          "name": {
            "type": [
              "string",
              "null"
            ]
          },
          "created_at": {
            "type": [
              "string",
              "null"
            ],
            "format": "date-time"
          },
          "active": {
            "type": [
              "boolean",
              "null"
            ]
          },
          "balance": {
            "type": [
              "number",
              "null"
            ]
          }
        }
      }
    }
  ]
}
"""


@pytest.fixture
def select_all_executor():
    return SelectExecutor(["*.*"])


@pytest.mark.parametrize(
    "path,prop",
    [
        ("stream[0].properties.master.properties.details", "master.details"),
        ("stream[2].properties.name", "name"),
        ("stream[10].properties.list[2].properties.name", "list[2].name"),
    ],
)
def test_path_property(path, prop):
    assert path_property(path) == prop


class TestSingerTap:
    @pytest.fixture(scope="class")
    def subject(self, project_add_service):
        return project_add_service.add(PluginType.EXTRACTORS, "tap-mock")

    def config_files(self, subject, dir: Path):
        return {key: dir.join(file) for key, file in subject.config_files.items()}

    def test_exec_args(self, subject, tmpdir):
        base_files = self.config_files(subject, tmpdir.mkdir("base"))
        assert subject.exec_args(base_files) == ["--config", base_files["config"]]

        # when `catalog` has data
        base_files = self.config_files(subject, tmpdir.mkdir("catalog"))
        base_files["catalog"].open("w").write("...")
        assert subject.exec_args(base_files) == [
            "--config",
            base_files["config"],
            "--catalog",
            base_files["catalog"],
        ]

        # when `state` has data
        base_files = self.config_files(subject, tmpdir.mkdir("state"))
        base_files["state"].open("w").write("...")
        assert subject.exec_args(base_files) == [
            "--config",
            base_files["config"],
            "--state",
            base_files["state"],
        ]

    def test_run_discovery(self, project, subject):
        invoker = PluginInvoker(project, subject)
        invoker.prepare()

        def mock_discovery():
            invoker.files["catalog"].open("w").write(CATALOG)
            return 0

        process_mock = mock.Mock()
        process_mock.wait = mock_discovery

        with mock.patch.object(
            PluginInvoker, "invoke", return_value=process_mock
        ) as invoke:
            subject.run_discovery(invoker, [])

            assert invoke.called_with(["--discover"])

    def test_run_discovery_fails(self, project, subject):
        process_mock = mock.Mock()
        process_mock.wait.return_value = 1  # something went wrong

        invoker = PluginInvoker(project, subject)
        invoker.prepare()

        with mock.patch.object(
            PluginInvoker, "invoke", return_value=process_mock
        ) as invoke, pytest.raises(PluginExecutionError):
            subject.run_discovery(invoker, [])

            assert not invoker.files[
                "catalog"
            ].exists(), "Catalog should not be present."

    def test_apply_select_catalog_invalid(self, project, subject):
        process_mock = mock.Mock()
        process_mock.wait.return_value = 0

        invoker = PluginInvoker(project, subject)
        invoker.prepare()

        def corrupt_catalog(*_, **__):
            invoker.files["catalog"].open("w").write("this is invalid json")

            return process_mock

        with mock.patch.object(
            PluginInvoker, "invoke", side_effect=corrupt_catalog
        ) as invoke:
            subject.apply_select(invoker, [])

            assert not invoker.files[
                "catalog"
            ].exists(), "Catalog should not be present."


class TestLegacyCatalogSelectVisitor:
    @pytest.fixture
    def catalog(self):
        return json.loads(LEGACY_CATALOG)

    @classmethod
    def stream_is_selected(cls, stream):
        return stream.get("selected", False)

    @classmethod
    def metadata_is_selected(cls, metadata):
        inclusion = metadata.get("inclusion")
        if inclusion == "automatic":
            return True

        return metadata.get("selected", False)

    @classmethod
    def assert_catalog_is_selected(cls, catalog):
        streams = {stream["stream"]: stream for stream in catalog["streams"]}

        metadatas = {
            stream["stream"]: metadata
            for _, stream in streams.items()
            for metadata in stream["metadata"]
        }

        # all streams are selected
        for name, stream in streams.items():
            assert cls.stream_is_selected(stream), f"{name} is not selected."

        # all fields are selected
        for stream, metadata in metadatas.items():
            field_metadata = metadata["metadata"]
            assert cls.metadata_is_selected(
                field_metadata
            ), f"{stream}.{metadata['breadcrumb']} is not selected"

    def test_visit(self, catalog, select_all_executor):
        visit(catalog, select_all_executor)

        self.assert_catalog_is_selected(catalog)


class TestCatalogSelectVisitor(TestLegacyCatalogSelectVisitor):
    @pytest.fixture
    def catalog(self, request):
        return json.loads(globals()[request.param])

    @classmethod
    def stream_is_selected(cls, stream):
        try:
            stream_metadata = next(
                metadata
                for metadata in stream["metadata"]
                if len(metadata["breadcrumb"]) == 0
            )

            return cls.metadata_is_selected(stream_metadata["metadata"])
        except (KeyError, IndexError):
            return False

    @pytest.mark.parametrize(
        "catalog", ["CATALOG", "JSON_SCHEMA"], indirect=["catalog"]
    )
    def test_visit(self, catalog, select_all_executor):
        super().test_visit(catalog, select_all_executor)

    @pytest.mark.parametrize(
        "catalog", ["CATALOG", "JSON_SCHEMA"], indirect=["catalog"]
    )
    def test_select_all(self, catalog, select_all_executor):
        visit(catalog, select_all_executor)
        self.assert_catalog_is_selected(catalog)

        streams = {stream["stream"]: stream for stream in catalog["streams"]}
        stream_metadata = len(
            [
                metadata
                for stream in streams.values()
                for metadata in stream["metadata"]
                if len(metadata["breadcrumb"]) == 0
            ]
        )

        assert stream_metadata == 1, "Extraneous stream metadata"

    @pytest.mark.parametrize(
        "catalog,attrs",
        [
            ("CATALOG", {"id", "code", "name", "code", "created_at"}),
            ("JSON_SCHEMA", {"id", "code", "name", "balance", "created_at", "active"}),
        ],
        indirect=["catalog"],
    )
    def test_select(self, catalog, attrs):
        selector = SelectExecutor(["entities.name", "entities.code"])
        visit(catalog, selector)

        lister = ListSelectedExecutor()
        visit(catalog, lister)

        assert lister.selected_properties["entities"] == attrs

    @pytest.mark.parametrize(
        "catalog,attrs",
        [
            ("CATALOG", {"id", "balance", "created_at", "active"}),
            ("JSON_SCHEMA", {"id", "code", "name", "balance", "created_at", "active"}),
        ],
        indirect=["catalog"],
    )
    def test_select_negated(self, catalog, attrs):
        selector = SelectExecutor(["*.*", "!entities.code", "!entities.name"])
        visit(catalog, selector)

        lister = ListSelectedExecutor()
        visit(catalog, lister)

        assert lister.selected_properties["entities"] == attrs


class TestListExecutor:
    @pytest.fixture
    def catalog(self):
        return json.loads(CATALOG)

    def test_visit(self, catalog):
        executor = ListExecutor()
        visit(catalog, executor)

        assert dict(executor.properties) == {
            "entities": {"code", "name", "balance", "created_at", "id", "active"}
        }


class TestSingerTarget:
    @pytest.fixture
    def subject(self, project_add_service):
        return project_add_service.add(PluginType.LOADERS, "target-mock")

    def test_exec_args(self, subject):
        base_files = subject.config_files
        assert subject.exec_args(base_files) == ["--config", base_files["config"]]



import logging
import subprocess
from io import StringIO

from . import Runner
from meltano.core.project import Project
from meltano.core.dbt_service import DbtService


class DbtRunner(Runner):
    def __init__(self, project: Project, dbt_service: DbtService = None):
        self.project = project
        self.dbt_service = dbt_service or DbtService(project)

    def run(self, dry_run=False, models=None):
        self.dbt_service.deps()

        if models is not None:
            models = models.replace("-", "_")

        if dry_run:
            self.dbt_service.compile(models)
        else:
            self.dbt_service.run(models)



import os
import json
from os.path import join
from pathlib import Path

from meltano.core.m5o.m5o_collection_parser import (
    M5oCollectionParser,
    M5oCollectionParserTypes,
)
from meltano.core.m5o.m5o_file_parser import MeltanoAnalysisFileParser
from meltano.core.project import Project
from meltano.core.utils import slugify
from .sql_helper import SqlHelper


class DashboardsHelper:
    VERSION = "1.0.0"

    def get_dashboards(self):
        project = Project.find()
        path = project.root_dir("model")
        dashboardsParser = M5oCollectionParser(path, M5oCollectionParserTypes.Dashboard)
        return dashboardsParser.contents()

    def get_dashboard_reports_with_query_results(self, reports):
        sqlHelper = SqlHelper()
        for report in reports:
            # TODO: refactor front-end `model` → `topic`
            m5oc = sqlHelper.get_m5oc_topic(report["model"])
            design = m5oc.design(report["design"])
            connection_name = m5oc.connection("connection")

            sql_dict = sqlHelper.get_sql(design, report["queryPayload"])
            outgoing_sql = sql_dict["sql"]
            aggregates = sql_dict["aggregates"]
            db_table = sql_dict["db_table"]

            report["queryResults"] = sqlHelper.get_query_results(
                connection_name, outgoing_sql
            )
            report["queryResultAggregates"] = aggregates
        return reports

    def get_dashboard(self, dashboard_id):
        dashboards = self.get_dashboards()
        target_dashboard = [
            dashboard for dashboard in dashboards if dashboard["id"] == dashboard_id
        ]
        return target_dashboard[0]

    def save_dashboard(self, data):
        project = Project.find()
        slug = slugify(data["name"])
        file_name = f"{slug}.dashboard.m5o"
        file_path = project.root_dir("model", file_name)
        data = MeltanoAnalysisFileParser.fill_base_m5o_dict(file_path, slug, data)
        data["version"] = DashboardsHelper.VERSION
        data["description"] = data["description"] or ""
        data["reportIds"] = []
        with open(file_path, "w") as f:
            json.dump(data, f)
        return data

    def add_report_to_dashboard(self, data):
        project = Project.find()
        dashboard = self.get_dashboard(data["dashboardId"])
        if data["reportId"] not in dashboard["reportIds"]:
            dashboard["reportIds"].append(data["reportId"])
            file_name = f"{dashboard['slug']}.dashboard.m5o"
            file_path = project.root_dir("model", file_name)
            with open(file_path, "w") as f:
                json.dump(dashboard, f)
        return dashboard

    def remove_report_from_dashboard(self, data):
        project = Project.find()
        dashboard = self.get_dashboard(data["dashboardId"])
        if data["reportId"] in dashboard["reportIds"]:
            dashboard["reportIds"].remove(data["reportId"])
            file_name = f"{dashboard['slug']}.dashboard.m5o"
            file_path = project.root_dir("model", file_name)
            with open(file_path, "w") as f:
                json.dump(dashboard, f)
        return dashboard
